import numpy as np
import os
import requests
from rasterio import features
from rasterio.windows import Window, from_bounds
from scgt import GeoTiff
from shapely import unary_union
from shapely.geometry import shape
from osgeo import gdal, ogr
from textwrap import dedent

from .redlist import RedList
from .utils import reproject_shapefile, make_dirs_for_file, get_current_habitat


class LayerGenerator:
    """Generate a Habitat Layer with various parameters to control the final output.
    The Habitat Layer is a raster that indicates what is considered habitat for a species.
    """

    def __init__(
        self,
        landcover_fn: str,
        redlist_key: str,
        ebird_key: str | None = None,
        elevation_fn: str | None = None,
        iucn_range_src: str | None = None,
    ):
        """Initializes a LayerGenerator object.

        Args:
            landcover_fn (str): file path to the initial landcover raster.
            redlist_key (str): IUCN Red List API key.
            ebird_key (str | None, optional): eBird API key. Defaults to None.
            elevation_fn (str | None, optional): file path to optional input elevation raster for filtering habitat by elevation; use None for no elevation consideration. Defaults to None.
            iucn_range_src (str | None, optional): file path to the IUCN range source if wanted. Defaults to None.
        """

        self.redlist = RedList(redlist_key, ebird_key)
        self.ebird_key = ebird_key
        self.landcover_fn = os.path.abspath(landcover_fn)
        self.elevation_fn = os.path.abspath(elevation_fn) if elevation_fn else None
        self.iucn_range_src = iucn_range_src

    def get_range_from_iucn(self, species_name: str, output_path: str):
        """Using IUCN gdb file, creates shapefiles usable for refining ranges for specific species with GDAL's ogr module.

        Args:
            species_name (str): scientific name of species to obtain range for.
            output_path (str): path for output .shp file (if it exists already, the old file(s) will be overwritten)
        """

        # We choose to use this option to avoid spending too much time organizing polygons.
        # See https://gdal.org/api/ogrgeometry_cpp.html#_CPPv4N18OGRGeometryFactory16organizePolygonsEPP11OGRGeometryiPiPPKc, https://gdal.org/user/configoptions.html#general-options.
        gdal.SetConfigOption("OGR_ORGANIZE_POLYGONS", "CCW_INNER_JUST_AFTER_CW_OUTER")

        # Open input file and layer, and apply attribute filter using scientific name
        input_src = ogr.Open(self.iucn_range_src, 0)
        input_layer = input_src.GetLayer()
        input_layer_defn = input_layer.GetLayerDefn()
        input_layer.SetAttributeFilter(f"sci_name = '{species_name}'")
        input_spatial_ref = input_layer.GetSpatialRef()
        input_spatial_ref.MorphToESRI()

        # Define output driver, delete old output file(s) if they exist
        output_driver = ogr.GetDriverByName("ESRI Shapefile")
        if os.path.exists(output_path):
            output_driver.DeleteDataSource(output_path)

        # Create the output shapefile
        output_src = output_driver.CreateDataSource(output_path)
        output_layer_name = os.path.splitext(os.path.split(output_path)[1])[0]
        output_layer = output_src.CreateLayer(
            output_layer_name, geom_type=ogr.wkbMultiPolygon
        )

        # Add fields to output
        for i in range(0, input_layer_defn.GetFieldCount()):
            output_layer.CreateField(input_layer_defn.GetFieldDefn(i))

        # Add filtered features to output
        for inFeature in input_layer:
            output_layer.CreateFeature(inFeature)

        # Create .prj file by taking the projection of the input file
        output_prj = open(os.path.splitext(output_path)[0] + ".prj", "w")
        output_prj.write(input_spatial_ref.ExportToWkt())
        output_prj.close()

        # Save and close files
        input_src = None
        output_src = None

        # Reset the GDAL config option
        gdal.SetConfigOption("OGR_ORGANIZE_POLYGONS", "DEFAULT")

    def get_range_from_ebird(self, species_code: str, output_path: str):
        """Gets range map in geopackage (.gpkg) format for a given bird species.

        Args:
            species_code (str): eBird code for a bird species.
            output_path (str): path to write the range map to.

        Raises:
            requests.HTTPError: Error occurs when the eBird API is expired or invalid
        """

        req_url = f"https://st-download.ebird.org/v1/fetch?objKey=2022/{species_code}/ranges/{species_code}_range_smooth_9km_2022.gpkg&key={self.ebird_key}"
        res = requests.get(req_url)

        if res.status_code == 200:
            with open(output_path, "wb") as res_file:
                res_file.write(res.content)
        else:
            error = f"""\
                Failed to download range map for {species_code} from eBird API.
                Your eBird API key may be expired or invalid."""
            raise requests.HTTPError(dedent(error))

    def generate_habitat(
        self,
        species_code: str,
        iucn_habitat_data: dict[int, dict[str, str | bool]] | None = None,
        habitat_fn: str | None = None,
        range_fn: str | None = None,
        range_src: str = "iucn",
        current_hab_overrides: list[str | int] | None = None,
    ):
        """Runner function for full process of habitat layer generation for one species.

        - iucn_habitat_data is an optional parameter, but should ideally be used to prevent unnecessarily fetching IUCN habitat data multiple times.
        - current_hab_overrides should most likely not be used to preserve consistency across what is determined as habitat.
        However, it gives users more custom control.

        Args:
            species_code (str): IUCN scientific name or eBird code. Determined based on value of range_src.
            iucn_habitat_data (dict[int, dict[str, int  |  float  |  str  |  bool]] | None, optional): The habitat data for a species received from IUCN Redlist. Defaults to None.
            habitat_fn (str | None, optional): The output path for the habitat layer. Defaults to None.
            range_fn (str | None, optional): The output path for the range map. This is created before making the habitat layer. Defaults to None.
            range_src (str, optional): The source from which to obtain range maps from. Defaults to "iucn".
            current_hab_overrides (list[str  |  int] | None, optional): This parameter is passed to the get_current_habitat function and allows the user to redefine what is considered habitat. Defaults to None.

        Raises:
            ValueError: If the user said they wanted eBird range maps but did not provide a ebird_key.
            ValueError: If the user said they wanted IUCN range maps but did not provide an IUCN range map data src.
            AssertionError: If the IUCN Redlist does not contain any habitat data for species.
            FileNotFoundError: If the range map file that is created/fetched cannot be found later in the habitat layer creation pipeline.
        """

        # Check for errors related to optional data provided to class
        if range_src == "ebird" and self.ebird_key is None:
            raise ValueError("eBird API key is required to get range maps from eBird.")

        if range_src == "iucn" and self.iucn_range_src is None:
            raise ValueError("iucn_range_src is required when range_src == 'iucn'.")

        # If file names not specified, build default ones.
        cwd = os.getcwd()
        if habitat_fn is None:
            habitat_fn = os.path.join(cwd, species_code, "habitat.tif")
        if range_fn is None:
            range_fn = os.path.join(cwd, species_code, "range_map.gpkg")

        # Ensure that directories to habitat layer, range map, and resistance dictionary exist.
        make_dirs_for_file(habitat_fn)
        make_dirs_for_file(range_fn)

        # Obtain scientific name of species if needed.
        if range_src == "iucn":
            sci_name = species_code
        else:
            sci_name = self.redlist.get_scientific_name(species_code)

        # Obtain species habitat information from the IUCN Red List.
        if iucn_habitat_data is None:
            iucn_habitat_data = self.redlist.get_habitat_data(sci_name)

        # Check for errors before processing habitat layer.
        if len(iucn_habitat_data) == 0:
            error = f"""\
                Habitat preferences for {species_code} could not be found on the IUCN Red List.
                Habitat layer and resistance dictionary were not generated."""
            raise AssertionError(dedent(error))

        if range_src == "iucn":
            self.get_range_from_iucn(sci_name, range_fn)
        else:
            self.get_range_from_ebird(species_code, range_fn)

        if not os.path.isfile(range_fn):
            src = "IUCN" if range_src == "iucn" else "eBird"
            error = f"Range map could not be found for {species_code} from {src}."
            raise FileNotFoundError(error)

        # Perform intersection between the range and habitable landcover.
        with GeoTiff.from_file(self.landcover_fn) as landcover:
            # Obtain species range as either shapefile from IUCN or geopackage from eBird.
            _, ext = os.path.splitext(range_fn)
            range_shapes = reproject_shapefile(
                range_fn, landcover.dataset.crs, "range" if ext == ".gpkg" else None
            )

            # Prepare range defined as shapes for masking
            if range_src == "iucn":
                for s in range_shapes:
                    if s["geometry"]["type"] == "Polygon":
                        s["geometry"]["coordinates"] = [
                            [el[0] for el in s["geometry"]["coordinates"][0]]
                        ]
                shapes_for_mask = [
                    unary_union([shape(s["geometry"]) for s in range_shapes])
                ]
            else:
                shapes_for_mask = [shape(range_shapes[0]["geometry"])]

            # Define map codes for which corresponding pixels should be considered habitat
            current_habitat_map_codes = get_current_habitat(
                iucn_habitat_data, current_hab_overrides
            )

            # Create the habitat layer
            with landcover.clone_shape(habitat_fn) as output:
                # If elevation raster is provided, obtain min/max elevation and read elevation raster
                if self.elevation_fn is not None:
                    min_elev, max_elev = self.redlist.get_elevation(sci_name)
                    elev = GeoTiff.from_file(self.elevation_fn)
                    cropped_window = (
                        from_bounds(
                            *output.dataset.bounds, transform=elev.dataset.transform
                        )
                        .round_lengths()
                        .round_offsets(pixel_precision=0)
                    )
                    x_offset, y_offset = cropped_window.col_off, cropped_window.row_off

                reader = output.get_reader(b=0, w=10000, h=10000)

                for tile in reader:
                    # get window and fit to the tiff's bounds if necessary
                    tile.fit_to_bounds(width=output.width, height=output.height)
                    window = Window(tile.x, tile.y, tile.w, tile.h)

                    # mask out pixels from landcover not within range of shapes
                    window_data = landcover.dataset.read(window=window, masked=True)
                    shape_mask = features.geometry_mask(
                        shapes_for_mask,
                        out_shape=(tile.h, tile.w),
                        transform=landcover.dataset.window_transform(window),
                    )
                    window_data.mask = window_data.mask | shape_mask
                    window_data = window_data.filled(0)

                    # get pixels where terrain is good
                    window_data = np.isin(window_data, current_habitat_map_codes)

                    # mask out pixels not within elevation range (if elevation raster is provided)
                    if self.elevation_fn is not None:
                        elev_window = Window(
                            tile.x + x_offset, tile.y + y_offset, tile.w, tile.h
                        )
                        elev_window_data = elev.dataset.read(window=elev_window)
                        window_data = (
                            window_data
                            & (elev_window_data >= min_elev)
                            & (elev_window_data <= max_elev)
                        )

                    # write the window result
                    output.dataset.write(window_data, window=window)

                if self.elevation_fn is not None:
                    elev.dataset.close()

            # This sets nodata to None for now, but should be changed later if scgt is modified to support that.
            with GeoTiff.from_file(habitat_fn) as output:
                output.dataset.nodata = None

        print("Habitat layer successfully generated for", species_code)



import fiona
import os
from pyproj import Transformer
import scgt
import numpy as np
from itertools import chain
from .constants import MAP_CODE_RANGES
import pandas as pd
from typing import Any, Callable
from textwrap import dedent
from osgeo import gdal
import tqdm
import psutil


def reproject_shapefile(
    shapes_path: str,
    dest_crs,
    shapes_layer: str | None = None,
    file_path: str | None = None,
) -> list[dict]:
    """
    Takes a specified shapefile or geopackage and reprojects it to a different CRS.

    :param shapes_path: file path to the shapefile or geopackage to reproject.
    :param dest_crs: CRS to reproject to as an ESRI WKT string.
    :param shapes_layer: if file is a geopackage, the name of the layer that should be reprojected.
    :param file_path: if specified, the file path to write the reprojected result to as a shapefile.
    :return: list of reprojected features.
    """

    features = []

    with fiona.open(shapes_path, "r", layer=shapes_layer) as shp:
        # create a Transformer for changing from the current CRS to the destination CRS
        transformer = Transformer.from_crs(
            crs_from=shp.crs_wkt, crs_to=dest_crs, always_xy=True
        )

        # loop through polygons in each features, transforming all point coordinates within those polygons
        for feature in shp:
            for i, polygon in enumerate(feature["geometry"]["coordinates"]):
                for j, ring in enumerate(polygon):
                    if isinstance(ring, list):
                        feature["geometry"]["coordinates"][i][j] = [
                            transformer.transform(*point) for point in ring
                        ]
                    else:
                        # "ring" is really just a single point
                        feature["geometry"]["coordinates"][i][j] = [
                            transformer.transform(*ring)
                        ]
            features.append(feature)

        # if file_path is specified, write the result to a new shapefile
        if file_path is not None:
            meta = shp.meta
            meta.update({"driver": "ESRI Shapefile", "crs_wkt": dest_crs})
            with fiona.open(file_path, "w", **meta) as output:
                output.writerecords(features)

    return features


def make_dirs_for_file(file_name: str):
    """Creates intermediate directories in the file path for a file if they don't exist yet.
    The file itself is not created; this just ensures that the directory of the file and all preceding ones
    exist first.

    Args:
        file_name (str): file to make directories for.
    """

    dirs, _ = os.path.split(file_name)
    os.makedirs(dirs, exist_ok=True)


def transform_box(
    min_lon: float,
    min_lat: float,
    max_lon: float,
    max_lat: float,
    crs_in: str,
    crs_out: str,
) -> tuple[float, float, float, float]:
    """
    Transforms a bounding box from one coordinate reference system (CRS) to another.

    Args:
        min_lon (float): The minimum longitude of the bounding box.
        min_lat (float): The minimum latitude of the bounding box.
        max_lon (float): The maximum longitude of the bounding box.
        max_lat (float): The maximum latitude of the bounding box.
        crs_in (str): The input CRS in EPSG format. For example "EPSG:4326".
        crs_out (str): The output CRS in EPSG format. For example "EPSG:3395".

    Returns:
        tuple: A tuple containing the transformed coordinates of the bounding box in the output CRS. The tuple has the format (min_x, min_y, max_x, max_y).
    """

    # Define in and out projections
    in_proj = Transformer.from_crs(crs_in, crs_out)

    # Transform corner points individually
    min_x, min_y = in_proj.transform(min_lat, min_lon)
    max_x, max_y = in_proj.transform(max_lat, max_lon)

    return min_x, min_y, max_x, max_y


def iucn_habs_to_codes(
    habitat_codes: list[str | int],
) -> list[str | int]:
    """Takes a list of strings and integers and replaces strings that are found in the IUCN habitat classification scheme with their respective map codes.
    This function does not guarantee removing all strings from the list, as there could be strings that are not found in the IUCN habitat classification scheme.

    Args:
        habitats (list[str  |  int]): A list of map codes which can contain strings matching the IUCN habitat classification scheme. Other strings can also be included but will not be replaced.

    Returns:
        list[str | int]: A list with all IUCN habitat classification scheme keywords replaced with map codes. Strings may still be present that are not found in IUCN habitat classification scheme.
    """

    for i in range(len(habitat_codes)):
        # check for map code range keyword from IUCN habitat classification scheme
        if habitat_codes[i] in MAP_CODE_RANGES:
            # get codes
            map_code_range = MAP_CODE_RANGES[str(habitat_codes[i])]
            new_codes = list(range(*map_code_range))

            # replace keyword with first code
            habitat_codes[i] = new_codes[0]

            # if new_codes have more than one code then put the rest at the end
            # NOTE: new_codes should have more than one code unless a keyword is made with one code
            if len(new_codes) > 1:
                habitat_codes.extend(new_codes[1:])

    return habitat_codes


def confirm_list_int(arr: list[Any]) -> list[int] | None:
    """Confirm that a list of unknown contents contains only integers.

    Args:
        arr (list[Any]): The list that is being checked.

    Returns:
        list[int] | None: A list of ints if successful and None if a non int type is found.
    """

    output: list[int] = []
    for elm in arr:
        if type(elm) is int:
            output.append(elm)
        else:
            # There was a non-int element
            return None

    return output


def in_habs(map_code: int, criteria: list[str | int]) -> bool:
    """Check if a map code is in a list of specified criteria. This criteria is made up of keywords that represent map code ranges and specific map codes.

    Examples:
        criteria_forest308: ["forest", 308]
        criteria_custom: ["forest", "shrubland", *list(range(600, 750))]

    Args:
        map_code (int): An integer map code.
        habitat_codes (list[str  |  int]): A list of strings and/or integers representing the criteria to check map_code against.

    Raises:
        KeyError: If the keyword used does not exist in the IUCN Habitat Classification Scheme.

    Returns:
        bool: True if a map code meets the criteria and False otherwise.
    """

    for criterion in criteria:
        if type(criterion) is str:
            if criterion in MAP_CODE_RANGES:
                rng = MAP_CODE_RANGES[criterion]
                if rng[0] <= map_code < rng[1]:
                    return True
            else:
                error = f"Keyword {criterion} not in IUCN Habitat Classification Scheme"
                raise KeyError(error)
        else:
            if map_code == criterion:
                return True

    return False


def get_map_codes(landcover: scgt.GeoTiff) -> list[int]:
    """Obtains the list of unique map codes present in the landcover/terrain map.

    Args:
        landcover (scgt.GeoTiff): The landcover/terrain map in GeoTiff format.

    Raises:
        ValueError: If the landcover/terrain map is empty.

    Returns:
        list[int]: The list of unique map codes.
    """

    # get landcover as a tile
    tile = landcover.get_all_as_tile()

    if tile is None:
        raise ValueError("Landcover file is empty.")

    # get map data from tile.m as a numpy matrix
    map_data: np.ndarray = tile.m

    # get unique map codes
    map_codes: list[int] = sorted(list(set(chain(map_data.flatten()))))

    return map_codes


def default_refinement_method(
    map_code: int, habitats: dict[int, dict[str, str | bool]]
) -> float:
    if map_code not in habitats:
        return 1.0

    if habitats[map_code]["majorimportance"]:
        return 0.0
    elif habitats[map_code]["suitability"]:
        return 0.1
    else:
        return 1.0


def generate_resistance_table(
    habitats: dict[int, dict[str, str | bool]],
    output_path: str,
    map_codes: list[int] = list(range(100, 1900)),
    refinement_method: Callable[
        [int, dict[int, dict[str, str | bool]]], float
    ] = default_refinement_method,
) -> pd.DataFrame:
    """Generates the resistance dictionary for a given species as a CSV file using habitat preference data from the IUCN Red List.

    - Link to map code definitions: https://www.iucnredlist.org/resources/habitat-classification-scheme
    - The default_refinement_method is a general use refinement method. If you want to use your own refinement method, the default_refinement_method is a great starting point.

    Args:
        habitats (dict[int, dict[str, str | bool]]): IUCN Red List habitat data for the species for which the table should be generated.
        output_path (str): The output path for the resistance csv.
        map_codes (list[int], optional): The map_codes that will be processed in the refinement method. The map codes defined in the habitat data are automatically added to this parameter. Default is all map codes from IUCN plus some extraneous map codes.
        refinement_method (function(int, dict[int, dict[str, str | bool]]) -> float, optional): Function that defines resistance when resistance is not defined in habitats. Defaults general pre-defined refinement method.

    Returns:
        DataFrame: The pandas data frame that is created into a csv.
    """

    # make sure the directory is created for resistance csv
    make_dirs_for_file(output_path)

    # make sure that habitats map codes are included in map_codes
    map_codes = list(set(map_codes + list(habitats.keys())))

    # create resistance data dict based on habitat data
    resistances: dict[int, dict[str, float | str | bool]] = {}
    for c, h in habitats.items():
        resistances[c] = {}
        for k, v in h.items():
            resistances[c][k] = v

    # Get resistance values for each map code
    for code in map_codes:
        # Create dict for code key in resistance dict
        if code not in resistances:
            resistances[code] = {}

        # use resistance from refinement method
        resistances[code]["resistance"] = refinement_method(code, habitats)

    # create a list of all dictionaries ordered on all IUCN codes
    dicts_ordered: list[dict[str, float | str | bool]] = []
    for code in range(100, 1900):
        if code in resistances:
            resistances[code]["map_code"] = code
            dicts_ordered.append(resistances[code])

    # Initialize a pandas data frame
    data = pd.DataFrame(dicts_ordered)

    # reorder the columns
    columns: list[str] = list(data.columns)

    # add the resistance to the end
    if "resistance" in columns:
        i = columns.index("resistance")
        if i != len(columns) - 1:
            columns.append(columns.pop(i))

    # add the map_code to the front
    if "map_code" in columns:
        i = columns.index("map_code")
        if i != 0:
            columns.insert(0, columns.pop(i))

    # set the data to new column order
    data = data[columns]

    # create the csv
    data.to_csv(output_path, index=False)

    return data


def get_current_habitat(
    habitats: dict[int, dict[str, str | bool]],
    overrides: list[str | int] | None = None,
) -> list[int]:
    """Determine the map codes that are considered habitat for a species.
    These map codes will then be used to create the habitat layer with habitat at these map codes.

    - Using the overrides parameter allows you to determine your custom definition of what should be considered habitat.
    Inputs for the overrides may consist of integers that represent specific map codes and strings that
    represent keywords, which will then be converted into map codes. These keywords can take the form of the
    IUCN Habitat Classification Scheme categories listed in the constants.py file. Additionally, you can
    specify "majorimportance" or "suitable" to only use habitats with these qualities for a species.

    Examples:
        overrides_forest308: ["forest", 308]
        overrides_custom: ["forest", "shrubland", *list(range(600, 750))]

    Args:
        habitats (dict[int, dict[str, int  |  float  |  str  |  bool]]): The habitat data received from the IUCN Redlist.
        overrides (list[str  |  int] | None, optional): List of strings/integers representing a custom definition of what constitutes habitat. Defaults to None.

    Raises:
        KeyError: This occurs when keyword is used in overrides that does not exist.
        AssertionError: This error should not occur, but if it does, there is an issue with the code implementation.

    Returns:
        list[int]: List of map codes that are considered habitat for a species.
    """

    # check if to use the override
    if overrides is not None:
        # remove all duplicates
        overrides = list(set(overrides))

        # replace all IUCN habitat classification scheme keywords with map codes
        overrides = iucn_habs_to_codes(overrides)

        # replace keywords (majorimportance, suitable) with map codes
        # search for the keywords and error on invalid keywords
        major_i = None
        suit_i = None
        for i in range(len(overrides)):
            if overrides[i] == "majorimportance":
                major_i = i
            elif overrides[i] == "suitable":
                suit_i = i
            elif type(overrides[i]) is str:
                error = f"""\
                    Keyword {overrides[i]} not found in IUCN habitat classification scheme keywords
                    and is not 'majorimportance' or 'suitable'."""
                raise KeyError(dedent(error))

        # define function to add new codes based on keyword
        def replace_keywords(keyword: str, keyword_i: int):
            if keyword_i is not None:
                new_codes: list[int] = []

                # search for codes that meet condition
                for code, hab in habitats.items():
                    if hab[keyword]:
                        new_codes.append(code)

                # check to suppress warnings
                if overrides is None:
                    return

                # replace keyword with first item
                overrides[keyword_i] = new_codes[0]

                # place remaining codes at the end
                if len(new_codes) > 1:
                    overrides.extend(new_codes[1:])

        # add new codes based on keywords
        if major_i is not None:
            replace_keywords("majorimportance", major_i)
        if suit_i is not None:
            replace_keywords("suitable", suit_i)

        # remove and duplicates due to overlap
        overrides = list(set(overrides))

        # check that all string have been removed
        output = confirm_list_int(overrides)
        if output is None:
            error = "There is an issue with the get_current_habitat implementation."
            raise AssertionError(error)

        return output

    # The default action is to return map_codes in which habitat is considered major importance by IUCN
    output = []
    for code, hab in habitats.items():
        if hab["majorimportance"]:
            output.append(code)

    return output


def warp(
    input: str,
    output: str,
    crs: str,
    resolution: float,
    bounds: tuple[float, float, float, float] | None = None,
    padding: int = 0,
    resampling: str = "near",
):
    """Transform a raster into another raster with various parameters.

    - Note that you cannot add padding to something that is not cropped with bounds, as otherwise there is no space for padding.

    Args:
        input (str): input file path for a raster
        output (str): output file path for a raster
        crs (str): output CRS
        resolution (float): x/y resolution
        bounds (tuple[float, float, float, float] | None, optional): output bounds in output CRS. Defaults to None.
        padding (int, optional): padding to add to the bounds. Defaults to 0.
        resampling (str, optional): resampling algorithm to use. See https://gdal.org/programs/gdalwarp.html#cmdoption-gdalwarp-r.
    """

    # Obtain input CRS
    input_src = gdal.Open(input, 0)
    input_crs = input_src.GetProjection()
    input_src = None

    if bounds is not None:
        padded_bounds = (
            bounds[0] - padding,
            bounds[1] - padding,
            bounds[2] + padding,
            bounds[3] + padding,
        )
    else:
        padded_bounds = None

    input_name = os.path.basename(input)
    progress = tqdm.tqdm(
        total=100, desc=f"Warping ({input_name})", unit="%", position=0
    )

    def _progress_callback(complete, message, data):
        progress.update(int(complete * 100 - progress.n))

    # get memory in bytes with minimum of 1GB or total memory
    # WARNING: gdal docs says that the memory max is in megabytes but it actual wants bytes from testing
    total_memory = psutil.virtual_memory().total
    min_memory = 1 * (1024**3)
    available_memory = psutil.virtual_memory().available
    if total_memory < min_memory:
        memory = None
    elif available_memory < min_memory:
        memory = min_memory
    else:
        memory = available_memory

    # Perform the warp using GDAL
    kwargs = {
        "format": "GTiff",
        "srcSRS": input_crs,
        "dstSRS": crs,
        "creationOptions": {
            "COMPRESS=LZW",
        },
        "outputBounds": padded_bounds,
        "xRes": resolution,
        "yRes": resolution,
        "resampleAlg": resampling,
        "multithread": True,
        "warpMemoryLimit": memory,
        "callback": _progress_callback,
    }

    gdal.Warp(output, input, **kwargs)

    progress.close()



from .layers import RedList, LayerGenerator
from .utils import (
    reproject_shapefile,
    make_dirs_for_file,
    transform_box,
    warp,
    iucn_habs_to_codes,
    get_map_codes,
    generate_resistance_table,
    get_current_habitat,
    confirm_list_int,
    in_habs,
    default_refinement_method,
)
from .constants import MAP_CODE_RANGES



RESAMPLING_METHODS = [
    "near",
    "bilinear",
    "cubic",
    "cubicspline",
    "lanczos",
    "average",
    "rms",
    "mode",
    "max",
    "min",
    "med",
    "q1",
    "q3",
    "sum",
]


# IUCN map code ranges
# NOTE: These codes can be found on https://www.iucnredlist.org/resources/habitat-classification-scheme
MAP_CODE_RANGES: dict[str, tuple[int, int]] = {
    "forest": (100, 200),
    "savanna": (200, 300),
    "shrubland": (300, 400),
    "grassland": (400, 500),
    "wetlands": (500, 600),
    "rocky": (600, 700),
    "caves": (700, 800),
    "desert": (800, 900),
    "marine": (900, 1400),
    "marine_neritic": (900, 1000),
    "marine_oceanic": (1000, 1100),
    "marine_deep": (1100, 1200),
    "marine_intertidal": (1200, 1300),
    "marine_coastal": (1300, 1400),
    "artificial": (1400, 1600),
    "artificial_terrestrial": (1400, 1500),
    "artificial_aquatic": (1500, 1600),
    "introduced_vegetation": (1600, 1700),
    "other": (1700, 1800),
    "unknown": (1800, 1900),
}



import requests
from ebird.api import get_taxonomy
from textwrap import dedent


class RedList:
    """
    A module of functions that primarily involve interfacing with the IUCN Red List API.
    """

    def __init__(self, redlist_key: str, ebird_key: str | None = None):
        """
        Initializes a RedList object.
        API keys are required to access the IUCN Red List API and eBird API respectively; see the documentation for more information.
        """
        self.redlist_params = {"token": redlist_key}
        self.ebird_key = ebird_key

    def get_from_redlist(self, url: str) -> dict:
        """
        Convenience function for sending GET request to Red List API with the key.

        :param url: the URL for the request.
        :return: response for the request.
        """
        res = requests.get(url, params=self.redlist_params)

        if res.status_code != 200:
            raise ValueError(f"Error {res.status_code} in Red List API request")

        data: dict = res.json()

        return data["result"]

    def get_scientific_name(self, species_code: str) -> str:
        """Translates eBird codes to scientific names for use in Red List.

        Args:
            species_code (str): eBird code for a bird species.

        Raises:
            ValueError: When the eBird API is not provided.
            ValueError: When the provided species_code is not found with eBird API

        Returns:
            str: The scientific name of the bird species
        """

        if self.ebird_key is None:
            raise ValueError("eBird API key is required to get scientific names")

        # Manual corrections here for differences between eBird and IUCN Red List scientific names.
        # This should probably be changed in the future.
        if species_code == "whhwoo":
            return "Leuconotopicus albolarvatus"
        elif species_code == "yebmag":
            return "Pica nutalli"
        elif species_code == "pilwoo":
            return "Hylatomus pileatus"
        elif species_code == "recwoo":
            return "Leuconotopicus borealis"

        res = get_taxonomy(self.ebird_key, species=species_code)
        sci_name = res[0]["sciName"] if len(res) > 0 else None

        if sci_name is None:
            error = f"sci_name for eBird code '{species_code}' could not be found."
            raise ValueError(error)

        return sci_name

    def get_habitat_data(
        self, species_name: str, region=None, ebird_code: bool = False
    ) -> dict[int, dict[str, str | bool]]:
        """Gets habitat assessments for suitability for a given species.
        This also adds the associated landcover/terrain map's code to the API response,
        which is useful for creating resistance mappings and/or habitat layers.
        Currently season data is ignored from IUCN habitat data.

        Args:
            species_name (str): scientific name of the species.
            region (_type_, optional): a specific region to assess habitats in (see https://apiv3.iucnredlist.org/api/v3/docs#regions).. Defaults to None.
            ebird_code (bool, optional): If True, reads species_name as an eBird species_code and converts it to a scientific/iucn name. Defaults to False.

        Raises:
            ValueError: Errors when the code received from the IUCN Redlist is missing a period or data after a period.
            KeyError: If the IUCN Redlist has two types of habitat with the

        Returns:
            dict[int, dict[str, str | bool]]: a dictionary map codes which reference habitats identified by the IUCN Red List as suitable for the species.
        """

        # Check if species_name needs to be converted to sci_name in the case of using ebird's species_code
        if ebird_code:
            sci_name = self.get_scientific_name(species_name)
        else:
            sci_name = species_name

        url = f"https://apiv3.iucnredlist.org/api/v3/habitats/species/name/{sci_name}"
        if region is not None:
            url += f"/region/{region}"

        habs = self.get_from_redlist(url)
        res = {}

        for hab in habs:
            code = str(hab["code"])

            # some codes are in the format xx.xx.xx instead of xx.xx
            # we will truncate xx.xx.xx codes to xx.xx
            code_sep = code.split(".")

            # check that code_sep len is not less than len of 2
            if len(code_sep) < 2:
                error = f"The code '{code}' is missing the required number of '.'"
                raise ValueError(error)

            # store bool to determine if code will be truncated
            truncated = len(code_sep) > 2

            # take a sub-array of the first two elements
            code_sep = code_sep[:2]

            # fill each number in two have length two so format is [xx, xx]
            code_sep = map(lambda num_str: num_str.zfill(2), code_sep)

            # Convert bool like strings to bools
            hab["majorimportance"] = hab["majorimportance"] == "Yes"
            hab["suitability"] = hab["suitability"] == "Suitable"

            # create a map_code that is represented by an int
            map_code = int("".join(code_sep))

            # add habitat to final result
            if map_code in res:
                # the map_code can be found multiple times if it is duplicate
                dup = True
                for k, v in res[map_code].items():
                    # dont check the season key has it is not handled currently
                    if k == "season":
                        continue

                    # check all other keys
                    if k not in hab or hab[k] != v:
                        dup = False
                        break
                if dup:
                    continue

                # The map_code can be found multiple times if the codes are truncated
                if truncated:
                    error = f"""\
                        Warning: A truncated map_code '{map_code}' was found multiple times:
                            first: {res[map_code]}
                            new  : {hab}
                        Using first data as interpreting multiple truncated map_codes is not currently supported.
                        """
                    print(dedent(error))
                    continue

                # Otherwise IUCN reports the same map_code twice
                error = f"""\
                    Warning: The map_code '{map_code}' was found multiple times:
                        first: {res[map_code]}
                        new  : {hab}
                    Using first data as IUCN redefining a map_code is unhandled.
                    """
                print(dedent(error))
            else:
                res[map_code] = hab

        return res

    def get_elevation(self, name) -> tuple[int, int]:
        """
        Obtain elevation bounds that are suitable for a given species
        :param name: scientific name of the species
        """
        url = f"https://apiv3.iucnredlist.org/api/v3/species/{name}"
        res = self.get_from_redlist(url)

        if len(res) == 0:
            return -10000, 10000
        else:
            # if elevation_lower is None, assume -10000; if elevation_upper is None, assume 10000
            return (
                int(res[0]["elevation_lower"]) or -10000,
                int(res[0]["elevation_upper"]) or 10000,
            )
